"""
Using multiple annotations and disabling draggability via signals
=================================================================

By default, each `Cursor` will ever display one annotation at a time.  Pass
``multiple=True`` to display multiple annotations.

Annotations can be made non-draggable by hooking their creation.
"""

import matplotlib.pyplot as plt
import numpy as np
import mplcursors

data = np.outer(range(10), range(1, 5))

fig, ax = plt.subplots()
ax.set_title("Multiple non-draggable annotations")
ax.plot(data)

mplcursors.cursor(multiple=True).connect(
    "add", lambda sel: sel.annotation.draggable(False))

plt.show()



"""
Annotate on hover
=================

When *hover* is set to ``True``, annotations are displayed when the mouse
hovers over the artist, without the need for clicking.
"""

import matplotlib.pyplot as plt
import numpy as np
import mplcursors
np.random.seed(42)

fig, ax = plt.subplots()
ax.scatter(*np.random.random((2, 26)))
ax.set_title("Mouse over a point")

mplcursors.cursor(hover=True)

plt.show()



"""
Display a bar's height and name on top of it upon hovering
==========================================================

Using an event handler to change the annotation text and position.
"""

import string
import matplotlib.pyplot as plt
import mplcursors

fig, ax = plt.subplots()
ax.bar(range(9), range(1, 10), align="center")
labels = string.ascii_uppercase[:9]
ax.set(xticks=range(9), xticklabels=labels, title="Hover over a bar")

# With HoverMode.Transient, the annotation is removed as soon as the mouse
# leaves the artist.  Alternatively, one can use HoverMode.Persistent (or True)
# which keeps the annotation until another artist gets selected.
cursor = mplcursors.cursor(hover=mplcursors.HoverMode.Transient)
@cursor.connect("add")
def on_add(sel):
    x, y, width, height = sel.artist[sel.index].get_bbox().bounds
    sel.annotation.set(text=f"{x+width/2}: {height}",
                       position=(0, 20), anncoords="offset points")
    sel.annotation.xy = (x + width / 2, y + height)

plt.show()



from pathlib import Path


def pytest_make_parametrize_id(config, val):
    if isinstance(val, type(lambda: None)) and val.__qualname__ != "<lambda>":
        return val.__qualname__
    if isinstance(val, Path):
        return str(val)



import json
import os

try:
    import importlib.metadata as _im
except ImportError:
    import importlib_metadata as _im
try:
    __version__ = _im.version("mplcursors")
except ImportError:
    __version__ = "0+unknown"

from ._mplcursors import Cursor, HoverMode, cursor
from ._pick_info import (
    Selection, compute_pick, get_ann_text, move, make_highlight)


__all__ = ["Cursor", "HoverMode", "cursor", "Selection",
           "compute_pick", "get_ann_text", "move", "make_highlight", "install"]


def install(figure):
    """
    A hook function that can be registered into ``rcParams["figure.hooks"]``.

    This hook arranges for a cursor to be registered on each figure the first
    time it is drawn, if the :envvar:`MPLCURSORS` environment variable is not
    empty (at first-draw time).  That variable must contain a JSON-encoded dict
    of options passed to `.cursor`.
    """

    def connect(event):
        figure.canvas.mpl_disconnect(cid)
        envopt = os.environ.get("MPLCURSORS")
        if not envopt:
            return
        cursor(figure, **json.loads(envopt))

    cid = figure.canvas.mpl_connect("draw_event", connect)



"""
Keyboard shortcuts
==================

By default, mplcursors uses "t" to toggle interactivity and "d" to hide/show
annotation boxes.  These shortcuts can be customized.
"""

import matplotlib.pyplot as plt
import mplcursors

fig, ax = plt.subplots()
ax.plot(range(10), "o-")
ax.set_title('Press "e" to enable/disable the datacursor\n'
             'Press "h" to hide/show any annotation boxes')

mplcursors.cursor(bindings={"toggle_visible": "h", "toggle_enabled": "e"})

plt.show()



# Unsupported Artist classes: subclasses of AxesImage, QuadMesh (upstream could
# have a `format_coord`-like method); PolyCollection (picking is not well
# defined).

from collections import namedtuple
from contextlib import suppress
import copy
import functools
import inspect
from inspect import Signature
from numbers import Integral
import re
import warnings
from weakref import WeakSet

from matplotlib.axes import Axes
from matplotlib.collections import (
    LineCollection, PatchCollection, PathCollection)
from matplotlib.container import BarContainer, ErrorbarContainer, StemContainer
from matplotlib.contour import ContourSet
from matplotlib.image import AxesImage, BboxImage
from matplotlib.lines import Line2D
from matplotlib.offsetbox import AnnotationBbox, OffsetBox
from matplotlib.patches import Patch, PathPatch, Polygon, Rectangle
from matplotlib.quiver import Barbs, Quiver
from matplotlib.text import Text
from matplotlib.transforms import (
    Affine2D, Bbox, BboxTransformFrom, BboxTransformTo)
import numpy as np


PATCH_PICKRADIUS = 5  # FIXME Patches do not provide `pickradius`.


def _register_scatter():
    """
    Patch `PathCollection` and `scatter` to register their return values.

    This registration allows us to distinguish `PathCollection`s created by
    `Axes.scatter`, which should use point-like picking, from others, which
    should use path-like picking.  The former is more common, so we store the
    latter instead; this also lets us guess the type better if this module is
    imported late.
    """

    @functools.wraps(PathCollection.__init__)
    def __init__(self, *args, **kwargs):
        _nonscatter_pathcollections.add(self)
        return __init__.__wrapped__(self, *args, **kwargs)
    PathCollection.__init__ = __init__

    @functools.wraps(Axes.scatter)
    def scatter(*args, **kwargs):
        paths = scatter.__wrapped__(*args, **kwargs)
        with suppress(KeyError):
            _nonscatter_pathcollections.remove(paths)
        return paths
    Axes.scatter = scatter


_nonscatter_pathcollections = WeakSet()
_register_scatter()


def _is_scatter(artist):
    return (isinstance(artist, PathCollection)
            and artist not in _nonscatter_pathcollections)


def _artist_in_container(container):
    return next(filter(None, container.get_children()))


class ContainerArtist:
    """Workaround to make containers behave more like artists."""

    def __init__(self, container):
        self.container = container  # Guaranteed to be nonempty.
        # We can't weakref the Container (which subclasses tuple), so
        # we instead create a reference cycle between the Container and
        # the ContainerArtist; as no one else strongly references the
        # ContainerArtist, it will get GC'd whenever the Container is.
        vars(container).setdefault(
            f"_{__class__.__name__}__keep_alive", []).append(self)

    def __str__(self):
        return f"<{type(self).__name__}({self.container})>"

    def __repr__(self):
        return f"<{type(self).__name__}({self.container!r})>"

    figure = property(lambda self: _artist_in_container(self.container).figure)
    axes = property(lambda self: _artist_in_container(self.container).axes)

    def get_visible(self):
        return True  # For lack of anything better.


Selection = namedtuple(
    "Selection", "artist target index dist annotation extras")
Selection.__doc__ = """
    A selection.

    Although this class is implemented as a namedtuple (to simplify the
    dispatching of `compute_pick`, `get_ann_text`, and `make_highlight`), only
    the field names should be considered stable API.  The number and order of
    fields are subject to change with no notice.
"""
# Override equality to identity: Selections should be considered immutable
# (with mutable fields though) and we don't want to trigger casts of array
# equality checks to booleans.  We don't need to override comparisons because
# artists are already non-comparable.
Selection.__eq__ = lambda self, other: self is other
Selection.__ne__ = lambda self, other: self is not other
Selection.artist.__doc__ = (
    "The selected artist.")
Selection.target.__doc__ = (
    "The point picked within the artist, in data coordinates.")
Selection.index.__doc__ = (
    "The index of the selected point, within the artist data.")
Selection.dist.__doc__ = (
    "The distance from the click to the target, in pixels.")
Selection.annotation.__doc__ = (
    "The instantiated `matplotlib.text.Annotation`.")
Selection.extras.__doc__ = (
    "An additional list of artists (e.g., highlighters) that will be cleared "
    "at the same time as the annotation.")


def _gen_warning_text(kind, tp):
    return "{} support for {} (MRO: {}) is missing.".format(
        kind, tp.__name__, ", ".join(cls.__name__ for cls in tp.__mro__))


@functools.singledispatch
def compute_pick(artist, event):
    """
    Find whether *artist* has been picked by *event*.

    If it has, return the appropriate `Selection`; otherwise return ``None``.

    This is a single-dispatch function; implementations for various artist
    classes follow.
    """
    warnings.warn(_gen_warning_text("Pick", type(artist)))


class Index:
    def __init__(self, i, x, y):
        self.int = i
        self.x = x
        self.y = y

    def __floor__(self):
        return self.int

    def __ceil__(self):
        return self.int if max(self.x, self.y) == 0 else self.int + 1

    # numpy<1.17 backcompat.
    floor = __floor__
    ceil = __ceil__

    def __format__(self, fmt):
        return f"{self.int}.(x={self.x:{fmt}}, y={self.y:{fmt}})"

    def __str__(self):
        return format(self, "")

    @classmethod
    def pre_index(cls, n_pts, index):
        i, frac = divmod(index, 1)
        i, odd = divmod(i, 2)
        x, y = (0, frac) if not odd else (frac, 1)
        return cls(i, x, y)

    @classmethod
    def post_index(cls, n_pts, index):
        i, frac = divmod(index, 1)
        i, odd = divmod(i, 2)
        x, y = (frac, 0) if not odd else (1, frac)
        return cls(i, x, y)

    @classmethod
    def mid_index(cls, n_pts, index):
        i, frac = divmod(index, 1)
        if i == 0:
            frac = .5 + frac / 2
        elif i == 2 * n_pts - 2:  # One less line than points.
            frac = frac / 2
        quot, odd = divmod(i, 2)
        if not odd:
            if frac < .5:
                i = quot - 1
                x, y = frac + .5, 1
            else:
                i = quot
                x, y = frac - .5, 0
        else:
            i = quot
            x, y = .5, frac
        return cls(i, x, y)


def _compute_projection_pick(artist, path, xy):
    """
    Project *xy* on *path* to obtain a `Selection` for *artist*.

    *path* is first transformed to screen coordinates using the artist
    transform, and the target of the returned `Selection` is transformed
    back to data coordinates using the artist *axes* inverse transform.  The
    `Selection` `index` is returned as a float.  This function returns ``None``
    for degenerate inputs.

    The caller is responsible for converting the index to the proper class if
    needed.
    """
    transform = artist.get_transform().frozen()
    tpath = (path.cleaned(transform) if transform.is_affine
             # `cleaned` only handles affine transforms.
             else transform.transform_path(path).cleaned())
    # `cleaned` should return a path where the first element is `MOVETO`, the
    # following are `LINETO` or `CLOSEPOLY`, and the last one is `STOP`, i.e.
    #     codes = path.codes
    #     assert (codes[0], codes[-1]) == (path.MOVETO, path.STOP)
    #     assert np.isin(codes[1:-1], [path.LINETO, path.CLOSEPOLY]).all()
    vertices = tpath.vertices[:-1]
    codes = tpath.codes[:-1]
    mt_idxs, = (codes == tpath.MOVETO).nonzero()
    cp_idxs, = (codes == tpath.CLOSEPOLY).nonzero()
    vertices[cp_idxs] = vertices[mt_idxs[mt_idxs.searchsorted(cp_idxs) - 1]]
    # Unit vectors for each segment.
    us = vertices[1:] - vertices[:-1]
    ls = np.hypot(*us.T)
    with np.errstate(invalid="ignore"):
        # Results in 0/0 for repeated consecutive points.
        us /= ls[:, None]
    # Vectors from each vertex to the event.
    vs = xy - vertices[:-1]
    # Clipped dot products -- `einsum` cannot be done in place, `clip` can.
    # `clip` can trigger invalid comparisons if there are nan points.
    with np.errstate(invalid="ignore"):
        dot = np.clip(np.einsum("ij,ij->i", vs, us), 0, ls)
    # Projections.
    projs = vertices[:-1] + dot[:, None] * us
    ds = np.hypot(*(xy - projs).T)
    ds[mt_idxs[1:] - 1] = np.nan
    try:
        argmin = np.nanargmin(ds)
    except ValueError:  # Raised by nanargmin([nan]).
        return
    else:
        target = artist.axes.transData.inverted().transform(projs[argmin])
        index = ((argmin + dot[argmin] / ls[argmin])
                 / (path._interpolation_steps / tpath._interpolation_steps))
        return Selection(artist, target, index, ds[argmin], None, None)


def _untransform(orig_xy, screen_xy, ax):
    """
    Return data coordinates to place an annotation at screen coordinates
    *screen_xy* in axes *ax*.

    *orig_xy* are the "original" coordinates as stored by the artist; they are
    transformed to *screen_xy* by whatever transform the artist uses.  If the
    artist uses ``ax.transData``, just return *orig_xy*; else, apply
    ``ax.transData.inverse()`` to *screen_xy*.  (The first case is more
    accurate than always applying ``ax.transData.inverse()``.)
    """
    tr_xy = ax.transData.transform(orig_xy)
    return (
        orig_xy
        if ((tr_xy == screen_xy) | np.isnan(tr_xy) & np.isnan(screen_xy)).all()
        else ax.transData.inverted().transform(screen_xy))


@compute_pick.register(Line2D)
def _(artist, event):
    # No need to call `line.contains` as we're going to redo the work anyways
    # (also see matplotlib/matplotlib#6645, though that's fixed in mpl2.1).

    # Always work in screen coordinates, as this is how we need to compute
    # distances.  Note that the artist transform may be different from the axes
    # transform (e.g., for axvline).
    xy = event.x, event.y
    data_xy = artist.get_xydata()
    data_screen_xy = artist.get_transform().transform(data_xy)
    sels = []
    # If markers are visible, find the closest vertex.
    if artist.get_marker() not in ["None", "none", " ", "", None]:
        ds = np.hypot(*(xy - data_screen_xy).T)
        try:
            argmin = np.nanargmin(ds)
        except ValueError:  # Raised by nanargmin([nan]).
            pass
        else:
            target = _untransform(  # More precise than transforming back.
                data_xy[argmin], data_screen_xy[argmin], artist.axes)
            sels.append(
                Selection(artist, target, argmin, ds[argmin], None, None))
    # If lines are visible, find the closest projection.
    if (artist.get_linestyle() not in ["None", "none", " ", "", None]
            and len(artist.get_xydata()) > 1):
        sel = _compute_projection_pick(artist, artist.get_path(), xy)
        if sel is not None:
            sel = sel._replace(index={
                "_draw_lines": lambda _, index: index,
                "_draw_steps_pre": Index.pre_index,
                "_draw_steps_mid": Index.mid_index,
                "_draw_steps_post": Index.post_index}[
                    Line2D.drawStyles[artist.get_drawstyle()]](
                        len(data_xy), sel.index))
            sels.append(sel)
    sel = min(sels, key=lambda sel: sel.dist, default=None)
    return sel if sel and sel.dist < artist.get_pickradius() else None


@compute_pick.register(PathPatch)
@compute_pick.register(Polygon)
@compute_pick.register(Rectangle)
def _(artist, event):
    sel = _compute_projection_pick(
        artist, artist.get_path(), (event.x, event.y))
    if sel and sel.dist < PATCH_PICKRADIUS:
        return sel


@compute_pick.register(LineCollection)
@compute_pick.register(PatchCollection)
@compute_pick.register(PathCollection)
def _(artist, event):
    offsets = artist.get_offsets()
    paths = artist.get_paths()
    if _is_scatter(artist):
        # Use the C implementation to prune the list of segments -- but only
        # for scatter plots as that implementation is inconsistent with Line2D
        # for segment-like collections (matplotlib/matplotlib#17279).
        contains, info = artist.contains(event)
        if not contains:
            return
        inds = info["ind"]
        offsets = artist.get_offsets()[inds]
        offsets_screen = artist.get_offset_transform().transform(offsets)
        ds = np.hypot(*(offsets_screen - [event.x, event.y]).T)
        argmin = ds.argmin()
        target = _untransform(
            offsets[argmin], offsets_screen[argmin], artist.axes)
        return Selection(artist, target, inds[argmin], ds[argmin], None, None)
    elif len(paths) and len(offsets):
        # Note that this won't select implicitly closed paths.
        sels = [
            _compute_projection_pick(
                artist,
                Affine2D().translate(*offsets[ind % len(offsets)])
                .transform_path(paths[ind % len(paths)]),
                (event.x, event.y))
            for ind in range(max(len(offsets), len(paths)))]
        if not any(sels):
            return None
        idx = min(range(len(sels)),
                  key=lambda idx: sels[idx].dist if sels[idx] else np.inf)
        sel = sels[idx]
        if sel.dist >= artist.get_pickradius():
            return None
        return sel._replace(artist=artist, index=(idx, sel.index))


# This registration has no effect on mpl<3.8, where ContourSets are not artists
# and thus do not appear in the draw tree.
# Filled contours are picked identically to unfilled ones, in particular
# because Path.contains_point does not handle holes in paths correctly; thus we
# cannot determine which contour, among many nested ones, actually contains the
# a point between two layers.
@compute_pick.register(ContourSet)
def _(artist, event):
    return compute_pick.dispatch(LineCollection)(artist, event)


@compute_pick.register(AxesImage)
@compute_pick.register(BboxImage)
def _(artist, event):
    if type(artist) not in compute_pick.registry:
        # Skip and warn on subclasses (`NonUniformImage`, `PcolorImage`) as
        # they do not implement `contains` correctly.  Even if they did, they
        # would not support moving as we do not know where a given index maps
        # back physically.
        return compute_pick.dispatch(object)(artist, event)
    contains, _ = artist.contains(event)
    if not contains:
        return
    ns = np.asarray(artist.get_array().shape[:2])[::-1]  # (y, x) -> (x, y)
    xf, yf = BboxTransformFrom(
        artist.get_window_extent()).transform([event.x, event.y])
    if artist.origin == "upper":
        yf = 1 - yf
    idxs = np.minimum(((xf, yf) * ns).astype(int), ns - 1)[::-1]
    return Selection(artist, (event.xdata, event.ydata),
                     tuple(idxs), 0, None, None)


@compute_pick.register(Barbs)
@compute_pick.register(Quiver)
def _(artist, event):
    offsets = artist.get_offsets()
    offsets_screen = artist.get_offset_transform().transform(offsets)
    ds = np.hypot(*(offsets_screen - [event.x, event.y]).T)
    argmin = np.nanargmin(ds)
    if ds[argmin] < artist.get_pickradius():
        target = _untransform(
            offsets[argmin], offsets_screen[argmin], artist.axes)
        return Selection(artist, target, argmin, ds[argmin], None, None)
    else:
        return None


@compute_pick.register(Text)
def _(artist, event):
    return


@compute_pick.register(ContainerArtist)
def _(artist, event):
    return compute_pick(artist.container, event)


@compute_pick.register(BarContainer)
def _(container, event):
    try:
        (idx, patch), = {
            (idx, patch) for idx, patch in enumerate(container.patches)
            if patch.contains(event)[0]}
    except ValueError:
        return
    target = [event.xdata, event.ydata]
    if patch.sticky_edges.x:
        target[0], = (
            x for x in [patch.get_x(), patch.get_x() + patch.get_width()]
            if x not in patch.sticky_edges.x)
    if patch.sticky_edges.y:
        target[1], = (
            y for y in [patch.get_y(), patch.get_y() + patch.get_height()]
            if y not in patch.sticky_edges.y)
    return Selection(container, target, idx, 0, None, None)


@compute_pick.register(ErrorbarContainer)
def _(container, event):
    data_line, cap_lines, err_lcs = container
    sel_data = compute_pick(data_line, event) if data_line else None
    sel_err = min(
        filter(None, (compute_pick(err_lc, event) for err_lc in err_lcs)),
        key=lambda sel: sel.dist, default=None)
    if (sel_data and sel_data.dist < getattr(sel_err, "dist", np.inf)):
        return sel_data
    elif sel_err:
        idx, _ = sel_err.index
        if data_line:
            target = data_line.get_xydata()[idx]
        else:  # We can't guess the original data in that case!
            return
        return Selection(container, target, idx, 0, None, None)
    else:
        return


@compute_pick.register(StemContainer)
def _(container, event):
    sel = compute_pick(container.markerline, event)
    if sel:
        return sel
    if not isinstance(container.stemlines, LineCollection):
        warnings.warn("Only stem plots created with use_line_collection=True "
                      "are supported.")
        return
    sel = compute_pick(container.stemlines, event)
    if sel:
        idx, _ = sel.index
        target = container.stemlines.get_segments()[idx][-1]
        return Selection(container, target, sel.index, 0, None, None)


@compute_pick.register(OffsetBox)
def _(artist, event):
    # Pass-through: actually picks a child artist.
    return min(
        filter(None, [compute_pick(child, event)
                      for child in artist.get_children()]),
        key=lambda sel: sel.dist, default=None)


@compute_pick.register(AnnotationBbox)
def _(artist, event):
    # Pass-through: actually picks a child artist.
    return compute_pick(artist.offsetbox, event)


def _call_with_selection(func=None, *, argname="artist"):
    """Decorator that passes a `Selection` built from the non-kwonly args."""

    if func is None:
        return functools.partial(_call_with_selection, argname=argname)

    wrapped_kwonly_params = [
        param for param in inspect.signature(func).parameters.values()
        if param.kind == param.KEYWORD_ONLY]
    sel_sig = inspect.signature(Selection)
    default_sel_sig = sel_sig.replace(
        parameters=[param.replace(default=None) if param.default is param.empty
                    else param
                    for param in sel_sig.parameters.values()])

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        extra_kw = {param.name: kwargs.pop(param.name)
                    for param in wrapped_kwonly_params if param.name in kwargs}
        ba = default_sel_sig.bind(*args, **kwargs)
        ba.apply_defaults()
        sel = Selection(*ba.args, **ba.kwargs)
        return func(sel, **extra_kw)

    params = [*sel_sig.parameters.values(), *wrapped_kwonly_params]
    params[0] = params[0].replace(name=argname)
    wrapper.__signature__ = Signature(params)
    return wrapper


def _format_coord_unspaced(ax, pos):
    # This used to directly post-process the output of format_coord(), but got
    # switched to handling special projections separately due to the change in
    # formatting for rectilinear coordinates.
    if ax.name == "polar":
        return ax.format_coord(*pos).replace(", ", "\n")
    elif ax.name == "3d":  # Need to retrieve the actual line data coordinates.
        warnings.warn("3d coordinates not supported yet")
        return ""
    else:
        x, y = pos
        # In mpl<3.3 (before #16776) format_x/ydata included trailing
        # spaces, hence the rstrip() calls.  format_xdata/format_ydata do not
        # actually always return strs (see test_fixed_ticks_nonstr_labels),
        # hence the explicit cast.
        return (f"x={str(ax.format_xdata(x)).rstrip()}\n"
                f"y={str(ax.format_ydata(y)).rstrip()}")


@functools.singledispatch
@_call_with_selection
def get_ann_text(sel):
    """
    Compute an annotating text for an (unpacked) `Selection`.

    This is a single-dispatch function; implementations for various artist
    classes follow.
    """
    warnings.warn(_gen_warning_text("Annotation", type(sel.artist)))
    return ""


@get_ann_text.register(Line2D)
@get_ann_text.register(LineCollection)
@get_ann_text.register(PatchCollection)
@get_ann_text.register(PathCollection)
@get_ann_text.register(Patch)
@_call_with_selection
def _(sel):
    artist = sel.artist
    label = artist.get_label() or ""
    text = _format_coord_unspaced(sel.annotation.axes, sel.target)
    if (_is_scatter(artist)
            # Heuristic: is the artist colormapped?
            # Note that this doesn't handle size-mapping (which is more likely
            # to involve an arbitrary scaling).
            and artist.get_array() is not None
            and len(artist.get_array()) == len(artist.get_offsets())):
        value = artist.format_cursor_data(artist.get_array()[sel.index])
        text = f"{text}\n{value}"
    if re.match("[^_]", label):
        text = f"{label}\n{text}"
    return text


@get_ann_text.register(ContourSet)
@_call_with_selection
def _(sel):
    artist = sel.artist
    return "{}\n{}".format(
        _format_coord_unspaced(sel.annotation.axes, sel.target),
        artist.levels[sel.index[0] + artist.filled])


@get_ann_text.register(AxesImage)
@get_ann_text.register(BboxImage)
@_call_with_selection
def _(sel):
    artist = sel.artist
    text = _format_coord_unspaced(sel.annotation.axes, sel.target)
    cursor_text = artist.format_cursor_data(artist.get_array()[sel.index])
    return f"{text}\n{cursor_text}"


@get_ann_text.register(Barbs)
@_call_with_selection
def _(sel):
    artist = sel.artist
    text = "{}\n({!s}, {!s})".format(
        _format_coord_unspaced(sel.annotation.axes, sel.target),
        artist.u[sel.index], artist.v[sel.index])
    return text


@get_ann_text.register(Quiver)
@_call_with_selection
def _(sel):
    artist = sel.artist
    text = "{}\n({!s}, {!s})".format(
        _format_coord_unspaced(sel.annotation.axes, sel.target),
        artist.U[sel.index], artist.V[sel.index])
    return text


# NOTE: There is no get_ann_text(ContainerArtist) as the selection directly
# refers to the Container itself.


@get_ann_text.register(BarContainer)
@_call_with_selection(argname="container")
def _(sel):
    return _format_coord_unspaced(sel.annotation.axes, sel.target)


@get_ann_text.register(ErrorbarContainer)
@_call_with_selection(argname="container")
def _(sel):
    data_line, cap_lines, err_lcs = sel.artist
    ann_text = get_ann_text(*sel._replace(artist=data_line))
    if isinstance(sel.index, Integral):
        err_lcs = iter(err_lcs)
        for idx, (dir, has) in enumerate(
                zip("xy", [sel.artist.has_xerr, sel.artist.has_yerr])):
            if has:
                err = (next(err_lcs).get_paths()[sel.index].vertices
                       - data_line.get_xydata()[sel.index])[:, idx]
                err_s = [getattr(sel.annotation.axes, f"format_{dir}data")(e)
                         .rstrip()
                         for e in err]
                # We'd normally want to check err.sum() == 0, but that can run
                # into fp inaccuracies.
                signs = "+-\N{MINUS SIGN}"
                if len({s.lstrip(signs) for s in err_s}) == 1:
                    repl = rf"\1=$\2\\pm{err_s[1]}$\3"
                else:
                    # Replacing unicode minus by ascii minus don't change the
                    # rendering as the string is mathtext, but allows keeping
                    # the same tests across Matplotlib versions that use
                    # unicode minus and those that don't.
                    err_s = [("+" if not s.startswith(tuple(signs)) else "")
                             + s.replace("\N{MINUS SIGN}", "-")
                             for s in err_s]
                    repl = r"\1=$\2_{%s}^{%s}$\3" % tuple(err_s)
                ann_text = re.sub(f"({dir})=(.*)(\n?)", repl, ann_text)
    return ann_text


@get_ann_text.register(StemContainer)
@_call_with_selection(argname="container")
def _(sel):
    return get_ann_text(*sel._replace(artist=sel.artist.markerline))


@functools.singledispatch
@_call_with_selection
def move(sel, *, key):
    """
    Move an (unpacked) `Selection` following a keypress.

    This function is used to implement annotation displacement through the
    keyboard.

    This is a single-dispatch function; implementations for various artist
    classes follow.
    """
    return sel


def _move_within_points(sel, xys, *, key):
    # Avoid infinite loop in case everything became nan at some point.
    for _ in range(len(xys)):
        if key == "left":
            new_idx = int(np.ceil(sel.index) - 1) % len(xys)
        elif key == "right":
            new_idx = int(np.floor(sel.index) + 1) % len(xys)
        else:
            return sel
        sel = sel._replace(target=xys[new_idx], index=new_idx, dist=0)
        if np.isfinite(sel.target).all():
            return sel


@move.register(Line2D)
@_call_with_selection
def _(sel, *, key):
    data_xy = sel.artist.get_xydata()
    return _move_within_points(
        sel,
        _untransform(data_xy, sel.artist.get_transform().transform(data_xy),
                     sel.annotation.axes),
        key=key)


@move.register(PathCollection)
@_call_with_selection
def _(sel, *, key):
    if _is_scatter(sel.artist):
        offsets = sel.artist.get_offsets()
        return _move_within_points(
            sel,
            _untransform(
                offsets, sel.artist.get_offset_transform().transform(offsets),
                sel.annotation.axes),
            key=key)
    else:
        return sel


@move.register(AxesImage)
@move.register(BboxImage)
@_call_with_selection
def _(sel, *, key):
    ns = sel.artist.get_array().shape[:2]
    delta = np.array(
        {"left": [0, -1], "right": [0, +1], "down": [-1, 0], "up": [+1, 0]}[
            key])
    if sel.artist.origin == "upper":
        delta[0] *= -1
    idxs = (sel.index + delta) % ns
    yf, xf = (idxs + .5) / ns
    if sel.artist.origin == "upper":
        yf = 1 - yf
    if isinstance(sel.artist, AxesImage):  # Same as below, but more accurate.
        x0, x1, y0, y1 = sel.artist.get_extent()
        trf = BboxTransformTo(Bbox.from_extents([x0, y0, x1, y1]))
    elif isinstance(sel.artist, BboxImage):
        trf = (BboxTransformTo(sel.artist.get_window_extent())
               - sel.annotation.axes.transData)
    target = trf.transform([xf, yf])
    return sel._replace(target=target, index=tuple(idxs))


# NOTE: There is no move(ContainerArtist) as the selection directly
# refers to the Container itself.


@move.register(ErrorbarContainer)
@_call_with_selection(argname="container")
def _(sel, *, key):
    data_line, cap_lines, err_lcs = sel.artist
    return _move_within_points(sel, data_line.get_xydata(), key=key)


@functools.singledispatch
@_call_with_selection
def make_highlight(sel, *, highlight_kwargs):
    """
    Create a highlight for an (unpacked) `Selection`.

    This is a single-dispatch function; implementations for various artist
    classes follow.
    """
    warnings.warn(_gen_warning_text("Highlight", type(sel.artist)))


def _set_valid_props(artist, kwargs):
    """Set valid properties for the artist, dropping the others."""
    artist.set(**{k: kwargs[k] for k in kwargs if hasattr(artist, "set_" + k)})
    return artist


@make_highlight.register(Line2D)
@_call_with_selection
def _(sel, *, highlight_kwargs):
    hl = copy.copy(sel.artist)
    _set_valid_props(hl, highlight_kwargs)
    return hl


@make_highlight.register(LineCollection)
@make_highlight.register(PathCollection)
@_call_with_selection
def _(sel, *, highlight_kwargs):
    hl = copy.copy(sel.artist)
    if _is_scatter(sel.artist):
        offsets = hl.get_offsets()
        hl.set_offsets(np.where(
            np.arange(len(offsets))[:, None] == sel.index, offsets, np.nan))
    else:
        hl.set_paths([
            path.vertices if i == sel.index[0] else np.empty((0, 2))
            for i, path in enumerate(hl.get_paths())])
    _set_valid_props(hl, highlight_kwargs)
    return hl



"""
Changing properties of the popup
================================

Use an event handler to customize the popup.
"""

import matplotlib.pyplot as plt
import numpy as np
import mplcursors

fig, axes = plt.subplots(ncols=2)

left_artist = axes[0].plot(range(11))
axes[0].set(title="No box, different position", aspect=1)

right_artist = axes[1].imshow(np.arange(100).reshape(10, 10))
axes[1].set(title="Fancy white background")

# Make the text pop up "underneath" the line and remove the box...
c1 = mplcursors.cursor(left_artist)
@c1.connect("add")
def _(sel):
    sel.annotation.set(position=(15, -15))
    # Note: Needs to be set separately due to matplotlib/matplotlib#8956.
    sel.annotation.set_bbox(None)

# Make the box have a white background with a fancier connecting arrow
c2 = mplcursors.cursor(right_artist)
@c2.connect("add")
def _(sel):
    sel.annotation.get_bbox_patch().set(fc="white")
    sel.annotation.arrow_patch.set(arrowstyle="simple", fc="white", alpha=.5)

plt.show()



from contextlib import ExitStack
import copy
import functools
import gc
import os
from pathlib import Path
import re
import subprocess
import sys
import weakref

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.backend_bases import KeyEvent, MouseEvent
import mplcursors
from mplcursors import _pick_info, Selection, HoverMode
import numpy as np
import pytest


# The absolute tolerance is quite large to take into account rounding of
# LocationEvents to the nearest pixel by Matplotlib, which causes a relative
# error of ~ 1/#pixels.
approx = functools.partial(pytest.approx, abs=1e-2)


@pytest.fixture
def fig():
    fig = plt.figure(1)
    fig.canvas.callbacks.exception_handler = None
    return fig


@pytest.fixture
def ax(fig):
    return fig.add_subplot(111)


@pytest.fixture(autouse=True)
def cleanup():
    with mpl.rc_context({"axes.unicode_minus": False}):
        try:
            yield
        finally:
            mplcursors.__warningregistry__ = {}
            plt.close("all")


def _process_event(name, ax, coords, *args):
    ax.viewLim  # unstale viewLim.
    if name == "__mouse_click__":
        # So that the dragging callbacks don't go crazy.
        _process_event("button_press_event", ax, coords, *args)
        _process_event("button_release_event", ax, coords, *args)
        return
    display_coords = ax.transData.transform(coords)
    if name in ["button_press_event", "button_release_event",
                "motion_notify_event", "scroll_event"]:
        event = MouseEvent(name, ax.figure.canvas, *display_coords, *args)
    elif name in ["key_press_event", "key_release_event"]:
        event = KeyEvent(name, ax.figure.canvas, *args, *display_coords)
    else:
        raise ValueError(f"Unknown event name {name!r}")
    ax.figure.canvas.callbacks.process(name, event)


def _get_remove_args(sel):
    ax = sel.artist.axes
    # Text bounds are found only upon drawing.
    ax.figure.canvas.draw()
    bbox = sel.annotation.get_window_extent()
    center = ax.transData.inverted().transform(
        ((bbox.x0 + bbox.x1) / 2, (bbox.y0 + bbox.y1) / 2))
    return "__mouse_click__", ax, center, 3


def _parse_annotation(sel, regex):
    result = re.fullmatch(regex, sel.annotation.get_text())
    assert result, \
        "{!r} doesn't match {!r}".format(sel.annotation.get_text(), regex)
    return tuple(map(float, result.groups()))


def test_containerartist(ax):
    artist = _pick_info.ContainerArtist(ax.errorbar([], []))
    str(artist)
    repr(artist)


def test_selection_identity_comparison():
    sel0, sel1 = [Selection(artist=None,
                            target=np.array([0, 0]),
                            index=None,
                            dist=0,
                            annotation=None,
                            extras=[])
                  for _ in range(2)]
    assert sel0 != sel1


def test_degenerate_inputs(ax):
    empty_container = ax.bar([], [])
    assert not mplcursors.cursor().artists
    assert not mplcursors.cursor(empty_container).artists
    pytest.raises(TypeError, mplcursors.cursor, [1])


@pytest.mark.parametrize("plotter", [Axes.plot, Axes.fill])
def test_line(ax, plotter):
    artist, = plotter(ax, [0, .2, 1], [0, .8, 1], label="foo")
    cursor = mplcursors.cursor(multiple=True)
    # Far, far away.
    _process_event("__mouse_click__", ax, (0, 1), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 0
    # On the line.
    _process_event("__mouse_click__", ax, (.1, .4), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 1
    assert _parse_annotation(
        cursor.selections[0], r"foo\nx=(.*)\ny=(.*)") == approx((.1, .4))
    # Not removing it.
    _process_event("__mouse_click__", ax, (0, 1), 3)
    assert len(cursor.selections) == len(ax.figure.artists) == 1
    # Remove the text label; add another annotation.
    artist.set_label(None)
    _process_event("__mouse_click__", ax, (.6, .9), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 2
    assert _parse_annotation(
        cursor.selections[1], r"x=(.*)\ny=(.*)") == approx((.6, .9))
    # Remove both of them (first removing the second one, to test
    # `Selection.__eq__` -- otherwise it is bypassed as `list.remove`
    # checks identity first).
    _process_event(*_get_remove_args(cursor.selections[1]))
    assert len(cursor.selections) == len(ax.figure.artists) == 1
    _process_event(*_get_remove_args(cursor.selections[0]))
    assert len(cursor.selections) == len(ax.figure.artists) == 0
    # Will project on the vertex at (.2, .8).
    _process_event("__mouse_click__", ax, (.2 - .001, .8 + .001), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 1


@pytest.mark.parametrize("plotter",
                         [lambda ax, *args: ax.plot(*args, ls="", marker="o"),
                          Axes.scatter])
def test_scatter(ax, plotter):
    plotter(ax, [0, .5, 1], [0, .5, 1])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.2, .2), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 0
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 1


def test_scatter_text(ax):
    ax.scatter([0, 1], [0, 1], c=[2, 3])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (0, 0), 1)
    assert _parse_annotation(
        cursor.selections[0], r"x=(.*)\ny=(.*)\n\[(.*)\]") == (0, 0, 2)


def test_steps_index():
    index = _pick_info.Index(0, .5, .5)
    assert np.floor(index) == 0 and np.ceil(index) == 1
    assert str(index) == "0.(x=0.5, y=0.5)"


def test_steps_pre(ax):
    ax.plot([0, 1], [0, 1], drawstyle="steps-pre")
    ax.set(xlim=(-1, 2), ylim=(-1, 2))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (1, 0), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (0, .5), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, 0, .5))
    _process_event("__mouse_click__", ax, (.5, 1), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, .5, 1))


def test_steps_mid(ax):
    ax.plot([0, 1], [0, 1], drawstyle="steps-mid")
    ax.set(xlim=(-1, 2), ylim=(-1, 2))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (0, 1), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (1, 0), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (.25, 0), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, .25, 0))
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, .5, .5))
    _process_event("__mouse_click__", ax, (.75, 1), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, .75, 1))


def test_steps_post(ax):
    ax.plot([0, 1], [0, 1], drawstyle="steps-post")
    ax.set(xlim=(-1, 2), ylim=(-1, 2))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (0, 1), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (.5, 0), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, .5, 0))
    _process_event("__mouse_click__", ax, (1, .5), 1)
    index = cursor.selections[0].index
    assert (index.int, index.x, index.y) == approx((0, 1, .5))


@pytest.mark.parametrize("ls", ["-", "o"])
def test_line_single_point(ax, ls):
    ax.plot(0, ls)
    ax.set(xlim=(-1, 1), ylim=(-1, 1))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.001, .001), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == (ls == "o")
    if cursor.selections:
        assert tuple(cursor.selections[0].target) == (0, 0)


@pytest.mark.parametrize("plot_args,click,targets",
                         [(([0, 1, np.nan, 3, 4],), (.5, .5), [(.5, .5)]),
                          (([np.nan, np.nan],), (0, 0), []),
                          (([np.nan, np.nan], "."), (0, 0), [])])
def test_nan(ax, plot_args, click, targets):
    ax.plot(*plot_args)
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, click, 1)
    assert len(cursor.selections) == len(ax.figure.artists) == len(targets)
    for sel, target in zip(cursor.selections, targets):
        assert sel.target == approx(target)


def test_repeated_point(ax):
    ax.plot([0, 1, 1, 2], [0, 1, 1, 2])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.5, .5), 1)  # Should not warn.


@pytest.mark.parametrize("origin", ["lower", "upper"])
@pytest.mark.parametrize("kind", ["AxesImage", "BboxImage"])
def test_image(ax, origin, kind):
    array = np.arange(6).reshape((3, 2))
    im = ax.imshow(array, origin=origin)  # Always set limits & orientation.

    if kind == "AxesImage":
        xzero = 0
    elif kind == "BboxImage":
        im.remove()
        ax.add_artist(
            mpl.image.BboxImage(
                mpl.transforms.TransformedBbox(
                    mpl.transforms.Bbox([[-.5, -.5], [1.5, 2.5]]), ax.transData),
                data=array, origin=origin))
        ax.figure.canvas.draw()  # Force image autonorm.
        xzero = approx(0)

    cursor = mplcursors.cursor()
    # Annotation text includes image value.
    _process_event("__mouse_click__", ax, (.25, .25), 1)
    sel, = cursor.selections
    assert _parse_annotation(  # Since mpl3.5 the value repr has extra zeros.
        sel, r"x=(.*)\ny=(.*)\n\[0(?:\.00)?\]") == approx((.25, .25))
    # Moving around.
    _process_event("key_press_event", ax, (.123, .456), "shift+right")
    sel, = cursor.selections
    assert _parse_annotation(sel, r"x=(.*)\ny=(.*)\n\[1(?:\.00)?\]") == (1, 0)
    assert array[sel.index] == 1
    _process_event("key_press_event", ax, (.123, .456), "shift+right")
    sel, = cursor.selections
    assert (_parse_annotation(sel, r"x=(.*)\ny=(.*)\n\[0(?:\.00)?\]")
            == (xzero, 0))
    assert array[sel.index] == 0
    _process_event("key_press_event", ax, (.123, .456), "shift+up")
    sel, = cursor.selections
    assert (_parse_annotation(sel, r"x=(.*)\ny=(.*)\n\[(.*)\]")
            == {"upper": (xzero, 2, 4), "lower": (0, 1, 2)}[origin])
    assert array[sel.index] == {"upper": 4, "lower": 2}[origin]
    _process_event("key_press_event", ax, (.123, .456), "shift+down")
    sel, = cursor.selections
    assert (_parse_annotation(sel, r"x=(.*)\ny=(.*)\n\[0(?:\.00)?\]")
            == (xzero, 0))
    assert array[sel.index] == 0

    cursor = mplcursors.cursor()
    # Not picking out of axes or out of image.
    _process_event("__mouse_click__", ax, (-1, -1), 1)
    assert len(cursor.selections) == 0
    ax.set(xlim=(-1, None), ylim=(-1, None))
    _process_event("__mouse_click__", ax, (-.75, -.75), 1)
    assert len(cursor.selections) == 0


def test_image_rgb(ax):
    ax.imshow([[[.1, .2, .3], [.4, .5, .6]]])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (0, 0), 1)
    sel, = cursor.selections
    assert _parse_annotation(
        sel, r"x=(.*)\ny=(.*)\n\[0.1, 0.2, 0.3\]") == approx((0, 0))
    _process_event("key_press_event", ax, (.123, .456), "shift+right")
    sel, = cursor.selections
    assert _parse_annotation(
        sel, r"x=(.*)\ny=(.*)\n\[0.4, 0.5, 0.6\]") == approx((1, 0))


def test_image_subclass(ax):
    # Cannot move around `PcolorImage`s.
    ax.pcolorfast(np.arange(3) ** 2, np.arange(3) ** 2, np.zeros((2, 2)))
    cursor = mplcursors.cursor()
    with pytest.warns(UserWarning):
        _process_event("__mouse_click__", ax, (1, 1), 1)
    assert len(cursor.selections) == 0


def test_linecollection(ax):
    ax.eventplot([])  # This must not raise a division by len([]) == 0.
    ax.eventplot([0, 1])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (0, 0), 1)
    _process_event("__mouse_click__", ax, (.5, 1), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (0, 1), 1)
    assert cursor.selections[0].index == approx((0, .5))


def test_patchcollection(ax):
    ax.add_collection(mpl.collections.PatchCollection([
        mpl.patches.Rectangle(xy, .1, .1) for xy in [(0, 0), (.5, .5)]]))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.05, .05), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (.6, .6), 1)
    # The precision is really bad :(
    assert cursor.selections[0].index == approx((1, 2), abs=2e-2)


@pytest.mark.skipif(getattr(mpl, "__version_info__", ()) < (3, 8),
                    reason="No support for old-style contours.")
def test_contour(ax):
    ax.contour([[0, 1], [0, 1]], levels=[.5])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    sel, = cursor.selections
    x, y, v = _parse_annotation(sel, r"x=(.*)\ny=(.*)\n(.*)")
    # The precision is really bad :(
    assert (x, y) == approx((.5, .5), abs=2e-2)
    assert v == .5


@pytest.mark.parametrize("plotter", [Axes.quiver, Axes.barbs])
def test_quiver_and_barbs(ax, plotter):
    plotter(ax, range(3), range(3))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.5, 0), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (1, 0), 1)
    assert _parse_annotation(
        cursor.selections[0], r"x=(.*)\ny=(.*)\n\(1, 1\)") == (1, 0)


@pytest.mark.parametrize("plotter,order",
                         [(Axes.bar, np.s_[:]), (Axes.barh, np.s_[::-1])])
def test_bar(ax, plotter, order):
    container = plotter(ax, range(3), range(1, 4))
    cursor = mplcursors.cursor()
    assert len(cursor.artists) == 1
    _process_event("__mouse_click__", ax, (0, 2)[order], 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (0, .5)[order], 1)
    assert cursor.selections[0].artist is container  # not the ContainerArtist.
    assert cursor.selections[0].target == approx((0, 1)[order])


def test_errorbar(ax):
    ax.errorbar(range(2), range(2), [(1, 1), (1, 2)])
    cursor = mplcursors.cursor()
    assert len(cursor.artists) == 1
    _process_event("__mouse_click__", ax, (0, 2), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert cursor.selections[0].target == approx((.5, .5))
    assert _parse_annotation(
        cursor.selections[0], r"x=(.*)\ny=(.*)") == approx((.5, .5))
    _process_event("__mouse_click__", ax, (0, 1), 1)
    assert cursor.selections[0].target == approx((0, 0))
    assert _parse_annotation(
        cursor.selections[0], r"x=(.*)\ny=\$(.*)\\pm(.*)\$") == (0, 0, 1)
    _process_event("__mouse_click__", ax, (1, 2), 1)
    sel, = cursor.selections
    assert sel.target == approx((1, 1))
    assert _parse_annotation(
        sel, r"x=(.*)\ny=\$(.*)_\{(.*)\}\^\{(.*)\}\$") == (1, 1, -1, 2)


def test_dataless_errorbar(ax):
    # Unfortunately, the original data cannot be recovered when fmt="none".
    ax.errorbar(range(2), range(2), [(1, 1), (1, 2)], fmt="none")
    cursor = mplcursors.cursor()
    assert len(cursor.artists) == 1
    _process_event("__mouse_click__", ax, (0, 0), 1)
    assert len(cursor.selections) == 0


def test_stem(ax):
    try:  # stem use_line_collection API change.
        ax.stem([1, 2, 3], use_line_collection=True)
    except TypeError:
        ax.stem([1, 2, 3])
    cursor = mplcursors.cursor()
    assert len(cursor.artists) == 1
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == 0
    _process_event("__mouse_click__", ax, (0, 1), 1)
    assert cursor.selections[0].target == approx((0, 1))
    _process_event("__mouse_click__", ax, (0, .5), 1)
    assert cursor.selections[0].target == approx((0, 1))


def test_annotationbbox(ax):
    ax.set(xlim=(0, 1), ylim=(0, 1))
    data = np.arange(9).reshape((3, 3))
    ax.add_artist(mpl.offsetbox.AnnotationBbox(
        mpl.offsetbox.OffsetImage(data, zoom=10), (.5, .5)))
    ax.figure.canvas.draw()
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    sel, = cursor.selections
    assert (_parse_annotation(sel, r"x=(.*)\ny=(.*)\n\[(.*)\]")
            == approx((.5, .5, 4)))


@pytest.mark.parametrize(
    "plotter,warns",
    [(lambda ax: ax.text(.5, .5, "foo"), False),
     (lambda ax: ax.fill_between([0, 1], [0, 1]), True)])
def test_misc_artists(ax, plotter, warns):
    plotter(ax)
    cursor = mplcursors.cursor()
    with pytest.warns(UserWarning) if warns else ExitStack():
        _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == 0


def test_indexless_projections(fig):
    ax = fig.subplots(subplot_kw={"projection": "polar"})
    ax.plot([1, 2], [3, 4])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (1, 3), 1)
    assert len(cursor.selections) == 1
    _process_event("key_press_event", ax, (.123, .456), "shift+left")


def test_cropped_by_axes(fig):
    axs = fig.subplots(2)
    axs[0].plot([0, 0], [0, 1])
    # Pan to hide the line behind the second axes.
    axs[0].set(xlim=(-1, 1), ylim=(1, 2))
    axs[1].set(xlim=(-1, 1), ylim=(-1, 1))
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", axs[1], (0, 0), 1)
    assert len(cursor.selections) == 0


@pytest.mark.parametrize("plotter", [Axes.plot, Axes.scatter, Axes.errorbar])
def test_move(ax, plotter):
    plotter(ax, [0, 1, 2], [0, 1, np.nan])
    cursor = mplcursors.cursor()
    # Nothing happens with no cursor.
    _process_event("key_press_event", ax, (.123, .456), "shift+left")
    assert len(cursor.selections) == 0
    # Now we move the cursor left or right.
    if plotter in [Axes.plot, Axes.errorbar]:
        _process_event("__mouse_click__", ax, (.5, .5), 1)
        assert tuple(cursor.selections[0].target) == approx((.5, .5))
        _process_event("key_press_event", ax, (.123, .456), "shift+up")
        _process_event("key_press_event", ax, (.123, .456), "shift+left")
    elif plotter is Axes.scatter:
        _process_event("__mouse_click__", ax, (0, 0), 1)
        _process_event("key_press_event", ax, (.123, .456), "shift+up")
    assert tuple(cursor.selections[0].target) == (0, 0)
    assert cursor.selections[0].index == 0
    _process_event("key_press_event", ax, (.123, .456), "shift+right")
    assert tuple(cursor.selections[0].target) == (1, 1)
    assert cursor.selections[0].index == 1
    # Skip through nan.
    _process_event("key_press_event", ax, (.123, .456), "shift+right")
    assert tuple(cursor.selections[0].target) == (0, 0)
    assert cursor.selections[0].index == 0


@pytest.mark.parametrize(
    "hover", [True, HoverMode.Persistent, 2, HoverMode.Transient])
def test_hover(ax, hover):
    l1, = ax.plot([0, 1])
    l2, = ax.plot([1, 2])
    cursor = mplcursors.cursor(hover=hover)
    _process_event("motion_notify_event", ax, (.5, .5), 1)
    assert len(cursor.selections) == 0  # No trigger if mouse button pressed.
    _process_event("motion_notify_event", ax, (.5, .5))
    assert cursor.selections[0].artist == l1
    _process_event("motion_notify_event", ax, (.5, 1))
    assert bool(cursor.selections) == (hover == HoverMode.Persistent)
    _process_event("motion_notify_event", ax, (.5, 1.5))
    assert cursor.selections[0].artist == l2


@pytest.mark.parametrize("plotter", [Axes.plot, Axes.scatter])
@pytest.mark.parametrize("remove_click_on_annotation", [True, False])
def test_highlight(ax, plotter, remove_click_on_annotation):
    plotter(ax, [0, 1], [0, 1])
    ax.set(xlim=(-1, 2), ylim=(-1, 2))
    base_children = {*ax.artists, *ax.lines, *ax.collections}
    cursor = mplcursors.cursor(highlight=True)
    _process_event("__mouse_click__", ax, (0, 0), 1)
    # On Matplotlib<=3.4, the highlight went to ax.artists.  On >=3.5, it goes
    # to its type-specific container.  The construct below handles both cases.
    assert [*{*ax.artists, *ax.lines, *ax.collections} - base_children] \
        == cursor.selections[0].extras != []
    if remove_click_on_annotation:
        _process_event(*_get_remove_args(cursor.selections[0]))
    else:
        _process_event("__mouse_click__", ax, (1, 1), 3)
    assert len({*ax.artists, *ax.lines, *ax.collections} - base_children) == (
        1 if plotter is Axes.scatter and not remove_click_on_annotation else 0)


def test_highlight_linecollection(ax):
    ax.eventplot([0, 1])
    cursor = mplcursors.cursor(highlight=True)
    _process_event("__mouse_click__", ax, (0, 1), 1)
    sel, = cursor.selections
    assert sel.extras


def test_misc_artists_highlight(ax):
    # Unsupported artists trigger a warning upon a highlighting attempt.
    ax.imshow([[0, 1], [2, 3]])
    cursor = mplcursors.cursor(highlight=True)
    with pytest.warns(UserWarning):
        _process_event("__mouse_click__", ax, (.5, .5), 1)


def test_callback(ax):
    ax.plot([0, 1])
    add_calls = []
    remove_calls = []
    cursor = mplcursors.cursor()
    on_add = cursor.connect("add")(lambda sel: add_calls.append(sel))
    on_remove = cursor.connect("remove")(lambda sel: remove_calls.append(sel))
    _process_event("__mouse_click__", ax, (.3, .3), 1)
    assert len(add_calls) == 1
    assert len(remove_calls) == 0
    _process_event("__mouse_click__", ax, (.7, .7), 1)
    assert len(add_calls) == 2
    assert len(remove_calls) == 1
    cursor.disconnect("add", on_add)
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(add_calls) == 2
    with pytest.raises(ValueError):
        cursor.disconnect("add", lambda sel: None)
    with pytest.raises(ValueError):
        cursor.connect("foo", lambda sel: None)


def test_remove_while_adding(ax):
    ax.plot([0, 1])
    cursor = mplcursors.cursor()
    cursor.connect("add", cursor.remove_selection)
    _process_event("__mouse_click__", ax, (.5, .5), 1)


def test_no_duplicate(ax):
    ax.plot([0, 1])
    cursor = mplcursors.cursor(multiple=True)
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == 1


def test_remove_multiple_overlapping(ax):
    ax.plot([0, 1])
    cursor = mplcursors.cursor(multiple=True)
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    sel, = cursor.selections
    cursor.add_selection(
        copy.copy(sel), sel.annotation.figure, sel.annotation.axes)
    assert len(cursor.selections) == 2
    _process_event(*_get_remove_args(sel))
    assert [*map(id, cursor.selections)] == [id(sel)]  # To check LIFOness.
    _process_event(*_get_remove_args(sel))
    assert len(cursor.selections) == 0


def test_autoalign(ax):
    ax.plot([0, 1])
    cursor = mplcursors.cursor()
    cb = cursor.connect(
        "add", lambda sel: sel.annotation.set(position=(-10, 0)))
    _process_event("__mouse_click__", ax, (.4, .4), 1)
    sel, = cursor.selections
    assert (sel.annotation.get_ha() == "right"
            and sel.annotation.get_va() == "center")
    cursor.disconnect("add", cb)
    cursor.connect(
        "add", lambda sel: sel.annotation.set(ha="center", va="bottom"))
    _process_event("__mouse_click__", ax, (.6, .6), 1)
    sel, = cursor.selections
    assert (sel.annotation.get_ha() == "center"
            and sel.annotation.get_va() == "bottom")


def test_drag(ax, capsys):
    l, = ax.plot([0, 1])
    cursor = mplcursors.cursor()
    cursor.connect(
        "add", lambda sel: sel.annotation.set(position=(0, 0)))
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    _process_event("button_press_event", ax, (.5, .5), 1)
    _process_event("motion_notify_event", ax, (.4, .6), 1)
    assert not capsys.readouterr().err


def test_removed_artist(ax):
    l, = ax.plot([0, 1])
    cursor = mplcursors.cursor()
    l.remove()
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 0


def test_remove_cursor(ax):
    ax.plot([0, 1])
    cursor = mplcursors.cursor()
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 1
    cursor.remove()
    assert len(cursor.selections) == len(ax.figure.artists) == 0
    _process_event("__mouse_click__", ax, (.5, .5), 1)
    assert len(cursor.selections) == len(ax.figure.artists) == 0


def test_keys(ax):
    ax.plot([0, 1])
    cursor = mplcursors.cursor(multiple=True)
    _process_event("__mouse_click__", ax, (.3, .3), 1)
    # Toggle visibility.
    _process_event("key_press_event", ax, (.123, .456), "v")
    assert not cursor.selections[0].annotation.get_visible()
    _process_event("key_press_event", ax, (.123, .456), "v")
    assert cursor.selections[0].annotation.get_visible()
    # Disable the cursor.
    _process_event("key_press_event", ax, (.123, .456), "e")
    assert not cursor.enabled
    # (Adding becomes inactive.)
    _process_event("__mouse_click__", ax, (.6, .6), 1)
    assert len(cursor.selections) == 1
    # (Removing becomes inactive.)
    ax.figure.canvas.draw()
    _process_event(*_get_remove_args(cursor.selections[0]))
    assert len(cursor.selections) == 1
    # (Moving becomes inactive.)
    old_target = cursor.selections[0].target
    _process_event("key_press_event", ax, (.123, .456), "shift+left")
    new_target = cursor.selections[0].target
    assert (old_target == new_target).all()
    # Reenable it.
    _process_event("key_press_event", ax, (.123, .456), "e")
    assert cursor.enabled
    _process_event(*_get_remove_args(cursor.selections[0]))
    assert len(cursor.selections) == 0


def test_select_at(ax):
    l1, = ax.plot([0, 1])
    l2, = ax.plot([1, 0])
    cursor = mplcursors.cursor([l1])
    assert cursor.select_at(l1, (.5, .5)) is not None
    assert len(cursor.selections) == 1
    cursor.remove_selection(cursor.selections[0])
    assert len(cursor.selections) == 0
    assert cursor.select_at(ax, (.2, .2)) is not None
    assert len(cursor.selections) == 1
    with pytest.raises(ValueError):
        cursor.select_at(l2, (.5, .5))


def test_convenience(ax):
    l, = ax.plot([1, 2])
    assert len(mplcursors.cursor().artists) == 1
    assert len(mplcursors.cursor(ax).artists) == 1
    assert len(mplcursors.cursor(l).artists) == 1
    assert len(mplcursors.cursor([l]).artists) == 1
    bc = ax.bar(range(3), range(3))
    assert len(mplcursors.cursor(bc).artists) == 1


@pytest.mark.skipif("figure.hooks" not in mpl.rcParams,
                    reason="Matplotlib version without figure.hooks.")
@pytest.mark.parametrize("envopt", ["", '{"hover": 1}'])
def test_figurehook(monkeypatch, envopt):
    monkeypatch.setenv("MPLCURSORS", envopt)
    with mpl.rc_context({"figure.hooks": ["mplcursors:install"]}):
        fig = plt.figure()
        try:
            ax = fig.add_subplot()
            ax.plot([0, 1])
            fig.canvas.draw()
            _process_event("motion_notify_event", ax, (.5, .5))
            assert len(fig.artists) == bool(envopt)  # The annotation.
        finally:
            plt.close(fig)


def test_invalid_args():
    pytest.raises(ValueError, mplcursors.cursor,
                  bindings={"foo": 42})
    pytest.raises(ValueError, mplcursors.cursor,
                  bindings={"select": 1, "deselect": 1})
    mplcursors.cursor(bindings={"select": None, "deselect": None})
    pytest.raises(ValueError, mplcursors.cursor().connect,
                  "foo")


def test_multiple_figures(ax):
    ax1 = ax
    _, ax2 = plt.subplots()
    ax1.plot([0, 1])
    ax2.plot([0, 1])
    cursor = mplcursors.cursor([ax1, ax2], multiple=True)
    # Add something on the first axes.
    _process_event("__mouse_click__", ax1, (.5, .5), 1)
    assert len(cursor.selections) == 1
    assert len(ax1.figure.artists) == 1
    assert len(ax2.figure.artists) == 0
    # Right-clicking on the second axis doesn't remove it.
    remove_args = [*_get_remove_args(cursor.selections[0])]
    remove_args[remove_args.index(ax1)] = ax2
    _process_event(*remove_args)
    assert len(cursor.selections) == 1
    assert len(ax1.figure.artists) == 1
    assert len(ax2.figure.artists) == 0
    # Remove it, add something on the second.
    _process_event(*_get_remove_args(cursor.selections[0]))
    _process_event("__mouse_click__", ax2, (.5, .5), 1)
    assert len(cursor.selections) == 1
    assert len(ax1.figure.artists) == 0
    assert len(ax2.figure.artists) == 1


def test_gc(ax):
    def inner():
        img = ax.imshow([[0, 1], [2, 3]])
        cursor = mplcursors.cursor(img)
        f_img = weakref.finalize(img, lambda: None)
        f_cursor = weakref.finalize(cursor, lambda: None)
        img.remove()
        return f_img, f_cursor
    f_img, f_cursor = inner()
    gc.collect()
    assert not f_img.alive
    assert not f_cursor.alive


def test_fixed_ticks_nonstr_labels(ax):
    ax.set_xticks([0])
    ax.set_xticklabels([0])  # The formatter will return the label as an int.
    ax.plot(0, 0, ".")
    cursor = mplcursors.cursor(ax)
    _process_event("__mouse_click__", ax, (0, 0), 1)
    # Just check that this does not error out, but don't check the annotation
    # text (it starts with "x=0" since #17266 but with "x=" before that).


@pytest.mark.parametrize(
    "example",
    [path for path in Path("examples").glob("*.py")
     if "test: skip" not in path.read_text()])
def test_example(example):
    subprocess.check_call(
        [sys.executable, "-mexamples.{}".format(example.with_suffix("").name)],
        # Unset $DISPLAY to avoid the non-GUI backend warning.
        env={**os.environ, "DISPLAY": "", "MPLBACKEND": "Agg"})



r"""
Contour plots
=============

Picking contour plots is supported on Matplotlib≥3.8.
"""

import numpy as np
import matplotlib.pyplot as plt
import mplcursors

fig, axs = plt.subplots(2, 2, figsize=(10, 4), sharex=True, sharey=True)

ii, jj = np.ogrid[:100, :100]
img = np.cos(ii / 20) * np.sin(jj / 10)
c = axs[0, 0].contour(img)
fig.colorbar(c, orientation="horizontal")
c = axs[0, 1].contourf(img)
fig.colorbar(c, orientation="horizontal")

ii, jj = np.random.rand(2, 1000) * 100
img = np.cos(ii / 20) * np.sin(jj / 10)
c = axs[1, 0].tricontour(jj, ii, img)
fig.colorbar(c, orientation="horizontal")
c = axs[1, 1].tricontourf(jj, ii, img)
fig.colorbar(c, orientation="horizontal")

mplcursors.cursor()

plt.show()



"""
Datetime data
=============

mplcursors correctly formats datetime data.
"""

import datetime as dt
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import mplcursors

t = mdates.drange(dt.datetime(2014, 1, 15), dt.datetime(2014, 2, 27),
                  dt.timedelta(hours=2))
y = np.sin(t)
fig, ax = plt.subplots()
ax.plot(mdates.num2date(t), y, "-")

# Note that mplcursors will automatically display the x-values as dates.
mplcursors.cursor()

plt.show()



"""
Scatter plots are highlighted point-by-point.
=============================================

... as opposed to lines with a ``"."`` style, which have the same appearance,
but are highlighted as a whole.
"""

import numpy as np
import matplotlib.pyplot as plt
import mplcursors

x, y, z = np.random.random((3, 10))
fig, axs = plt.subplots(3)
fig.suptitle("Highlighting affects individual points\n"
             "only in scatter plots (top two axes)")
axs[0].scatter(x, y, c=z, s=100 * np.random.random(10))
axs[1].scatter(x, y)
axs[2].plot(x, y, "o")
mplcursors.cursor(highlight=True)
plt.show()



"""
Highlighting the artist upon selection
======================================

Just pass ``highlight=True`` to `cursor`.
"""

import numpy as np
import matplotlib.pyplot as plt
import mplcursors

x = np.linspace(0, 10, 100)

fig, ax = plt.subplots()

# Plot a series of lines with increasing slopes.
lines = []
for i in range(1, 20):
    line, = ax.plot(x, i * x, label=f"$y = {i}x$")
    lines.append(line)

mplcursors.cursor(lines, highlight=True)

plt.show()



"""
Display an artist's label instead of x, y coordinates
=====================================================

Use an event handler to change the annotation text.
"""

import numpy as np
import matplotlib.pyplot as plt
import mplcursors

x = np.linspace(0, 10, 100)

fig, ax = plt.subplots()
ax.set_title("Click on a line to display its label")

# Plot a series of lines with increasing slopes.
for i in range(1, 20):
    ax.plot(x, i * x, label=f"$y = {i}x$")

# Use a Cursor to interactively display the label for a selected line.
mplcursors.cursor().connect(
    "add", lambda sel: sel.annotation.set_text(sel.artist.get_label()))

plt.show()



from sphinx_gallery.sorting import ExampleTitleSortKey


class CustomSortKey(ExampleTitleSortKey):
    def __call__(self, filename):
        return ("" if filename == "basic.py"  # goes first
                else super().__call__(filename))



"""
Extracting data and labels from a :class:`~pandas.DataFrame`
============================================================

:class:`~pandas.DataFrame`\\s can be used similarly to any other kind of input.
Here, we generate a scatter plot using two columns and label the points using
all columns.

This example also applies a shadow effect to the hover panel.
"""

from matplotlib import pyplot as plt

from matplotlib.patheffects import withSimplePatchShadow
import mplcursors
from pandas import DataFrame


df = DataFrame(
    dict(
        Suburb=["Ames", "Somerset", "Sawyer"],
        Area=[1023, 2093, 723],
        SalePrice=[507500, 647000, 546999],
    )
)

df.plot.scatter(x="Area", y="SalePrice", s=100)


def show_hover_panel(get_text_func=None):
    cursor = mplcursors.cursor(
        hover=2,  # Transient
        annotation_kwargs=dict(
            bbox=dict(
                boxstyle="square,pad=0.5",
                facecolor="white",
                edgecolor="#ddd",
                linewidth=0.5,
                path_effects=[withSimplePatchShadow(offset=(1.5, -1.5))],
            ),
            linespacing=1.5,
            arrowprops=None,
        ),
        highlight=True,
        highlight_kwargs=dict(linewidth=2),
    )

    if get_text_func:
        cursor.connect(
            event="add",
            func=lambda sel: sel.annotation.set_text(get_text_func(sel.index)),
        )
        
    return cursor


def on_add(index):
    item = df.iloc[index]
    parts = [
        f"Suburb: {item.Suburb}",
        f"Area: {item.Area:,.0f}m²",
        f"Sale price: ${item.SalePrice:,.0f}",
    ]

    return "\n".join(parts)


show_hover_panel(on_add)

plt.show()

# test: skip



import os
from pathlib import Path
import re
import sys
import mplcursors

# -- General configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_gallery.gen_gallery',
]

source_suffix = '.rst'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
master_doc = 'index'

project = 'mplcursors'
copyright = '2016–present, Antony Lee'
author = 'Antony Lee'

# RTD modifies conf.py, making setuptools_scm mark the version as -dirty.
version = release = re.sub(r'\.dirty$', '', mplcursors.__version__)

language = 'en'

default_role = 'any'

pygments_style = 'sphinx'

todo_include_todos = False

python_use_unqualified_type_names = True

# -- Options for HTML output ----------------------------------------------

html_theme = 'pydata_sphinx_theme'
html_theme_options = {
    'github_url': 'https://github.com/anntzer/mplcursors',
}
html_css_files = ['hide_some_gallery_elements.css']
html_static_path = ['_static']

htmlhelp_basename = 'mplcursors_doc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}
latex_documents = [(
    master_doc,
    'mplcursors.tex',
    'mplcursors Documentation',
    'Antony Lee',
    'manual',
)]

# -- Options for manual page output ---------------------------------------

man_pages = [(
    master_doc,
    'mplcursors',
    'mplcursors Documentation',
    [author],
    1,
)]

# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [(
    master_doc,
    'mplcursors',
    'mplcursors Documentation',
    author,
    'mplcursors',
    'Interactive data selection cursors for Matplotlib.',
    'Miscellaneous',
)]

# -- Misc. configuration --------------------------------------------------

autodoc_member_order = 'bysource'

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'matplotlib': ('https://matplotlib.org/stable', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
}

# CustomSortKey cannot be defined *here* because it would be unpicklable as
# this file is exec'd rather than imported.
sys.path.append(os.path.dirname(__file__))

os.environ.pop("DISPLAY", None)  # Don't warn about non-GUI when running s-g.

sphinx_gallery_conf = {
    'backreferences_dir': None,
    'examples_dirs': '../../examples',
    'filename_pattern': r'.*\.py',
    'gallery_dirs': 'examples',
    'min_reported_time': 1,
    'within_subsection_order': '_local_ext.CustomSortKey',
}



from collections.abc import Iterable
from contextlib import suppress
import copy
from enum import IntEnum
import functools
from functools import partial
import sys
import weakref
from weakref import WeakKeyDictionary, WeakSet

import matplotlib as mpl
from matplotlib.axes import Axes
from matplotlib.backend_bases import MouseEvent
from matplotlib.container import Container
from matplotlib.figure import Figure
import numpy as np

from . import _pick_info


_default_bindings = dict(
    select=1,
    deselect=3,
    left="shift+left",
    right="shift+right",
    up="shift+up",
    down="shift+down",
    toggle_enabled="e",
    toggle_visible="v",
)
_default_annotation_kwargs = dict(
    bbox=dict(
        boxstyle="round,pad=.5",
        fc="yellow",
        alpha=.5,
        ec="k",
    ),
    arrowprops=dict(
        arrowstyle="->",
        connectionstyle="arc3",
        shrinkB=0,
        ec="k",
    ),
)
_default_annotation_positions = [
    dict(position=(-15, 15), anncoords="offset points",
         horizontalalignment="right", verticalalignment="bottom"),
    dict(position=(15, 15), anncoords="offset points",
         horizontalalignment="left", verticalalignment="bottom"),
    dict(position=(15, -15), anncoords="offset points",
         horizontalalignment="left", verticalalignment="top"),
    dict(position=(-15, -15), anncoords="offset points",
         horizontalalignment="right", verticalalignment="top"),
]
_default_highlight_kwargs = dict(
    # Only the kwargs corresponding to properties of the artist will be passed.
    # Line2D.
    color="yellow",
    markeredgecolor="yellow",
    linewidth=3,
    markeredgewidth=3,
    # PathCollection.
    facecolor="yellow",
    edgecolor="yellow",
)


class _MarkedStr(str):
    """A string subclass solely for marking purposes."""


def _mouse_event_matches(event, spec):
    """
    Return whether a mouse event "matches" an event spec, which is either a
    single mouse button, or a mapping matched against ``vars(event)``, e.g.
    ``{"button": 1, "key": "control"}``.
    """
    if isinstance(spec, int):
        spec = {"button": spec}
    return all(getattr(event, k) == v for k, v in spec.items())


def _get_rounded_intersection_area(bbox_1, bbox_2):
    """Compute the intersection area between two bboxes rounded to 8 digits."""
    # The rounding allows sorting areas without floating point issues.
    bbox = bbox_1.intersection(bbox_1, bbox_2)
    return round(bbox.width * bbox.height, 8) if bbox else 0


def _iter_axes_subartists(ax):
    r"""Yield all child `Artist`\s (*not* `Container`\s) of *ax*."""
    yield from ax.collections
    yield from ax.images
    yield from ax.lines
    yield from ax.patches
    yield from ax.texts
    yield from ax.artists


def _is_alive(artist):
    """Check whether *artist* is still present on its parent axes."""
    return bool(
        artist
        and artist.axes
        # `cla()` clears `.axes` since matplotlib/matplotlib#24627 (3.7);
        # iterating over subartists can be very slow.
        and (getattr(mpl, "__version_info__", ()) >= (3, 7)
             or (artist.container in artist.axes.containers
                 if isinstance(artist, _pick_info.ContainerArtist) else
                 artist in _iter_axes_subartists(artist.axes))))


def _reassigned_axes_event(event, ax):
    """Reassign *event* to *ax*."""
    event = copy.copy(event)
    event.xdata, event.ydata = (
        ax.transData.inverted().transform((event.x, event.y)))
    return event


class HoverMode(IntEnum):
    NoHover, Persistent, Transient = range(3)


class Cursor:
    """
    A cursor for selecting Matplotlib artists.

    Attributes
    ----------
    bindings : dict
        See the *bindings* keyword argument to the constructor.
    annotation_kwargs : dict
        See the *annotation_kwargs* keyword argument to the constructor.
    annotation_positions : dict
        See the *annotation_positions* keyword argument to the constructor.
    highlight_kwargs : dict
        See the *highlight_kwargs* keyword argument to the constructor.
    """

    _keep_alive = WeakKeyDictionary()

    def __init__(self,
                 artists,
                 *,
                 multiple=False,
                 highlight=False,
                 hover=False,
                 bindings=None,
                 annotation_kwargs=None,
                 annotation_positions=None,
                 highlight_kwargs=None):
        """
        Construct a cursor.

        Parameters
        ----------

        artists : List[Artist]
            A list of artists that can be selected by this cursor.

        multiple : bool, default: False
            Whether multiple artists can be "on" at the same time.  If on,
            cursor dragging is disabled (so that one does not end up with many
            cursors on top of one another).

        highlight : bool, default: False
            Whether to also highlight the selected artist.  If so,
            "highlighter" artists will be placed as the first item in the
            :attr:`extras` attribute of the `Selection`.

        hover : `HoverMode`, default: False
            Whether to select artists upon hovering instead of by clicking.
            (Hovering over an artist while a button is pressed will not trigger
            a selection; right clicking on an annotation will still remove it.)
            Possible values are

            - False, alias `HoverMode.NoHover`: hovering is inactive.
            - True, alias `HoverMode.Persistent`: hovering is active;
              annotations remain in place even after the mouse moves away from
              the artist (until another artist is selected, if *multiple* is
              False).
            - 2, alias `HoverMode.Transient`: hovering is active; annotations
              are removed as soon as the mouse moves away from the artist.

        bindings : dict, optional
            A mapping of actions to button and keybindings.  Valid keys are:

            ================ ==================================================
            'select'         mouse button to select an artist
                             (default: :data:`.MouseButton.LEFT`)
            'deselect'       mouse button to deselect an artist
                             (default: :data:`.MouseButton.RIGHT`)
            'left'           move to the previous point in the selected path,
                             or to the left in the selected image
                             (default: shift+left)
            'right'          move to the next point in the selected path, or to
                             the right in the selected image
                             (default: shift+right)
            'up'             move up in the selected image
                             (default: shift+up)
            'down'           move down in the selected image
                             (default: shift+down)
            'toggle_enabled' toggle whether the cursor is active
                             (default: e)
            'toggle_visible' toggle default cursor visibility and apply it to
                             all cursors (default: v)
            ================ ==================================================

            Missing entries will be set to the defaults.  In order to not
            assign any binding to an action, set it to ``None``.  Modifier keys
            (or other event properties) can be set for mouse button bindings by
            passing them as e.g. ``{"button": 1, "key": "control"}``.

        annotation_kwargs : dict, default: {}
            Keyword argments passed to the `annotate
            <matplotlib.axes.Axes.annotate>` call.

        annotation_positions : List[dict], optional
            List of positions tried by the annotation positioning algorithm.
            The default is to try four positions, 15 points to the NW, NE, SE,
            and SW from the selected point; annotations that stay within the
            axes are preferred.

        highlight_kwargs : dict, default: {}
            Keyword arguments used to create a highlighted artist.
        """

        artists = [*artists]
        # Be careful with GC.
        self._artists = [weakref.ref(artist) for artist in artists]

        for artist in artists:
            type(self)._keep_alive.setdefault(artist, set()).add(self)

        self._multiple = multiple
        self._highlight = highlight

        self._visible = True
        self._enabled = True
        self._selections = []
        self._selection_stack = []
        self._last_auto_position = None
        self._callbacks = {"add": [], "remove": []}
        self._hover = hover

        self._suppressed_events = WeakSet()
        connect_pairs = [
            ("pick_event", self._on_pick),
            ("key_press_event", self._on_key_press),
        ]
        if hover:
            connect_pairs += [
                ("motion_notify_event", self._on_hover_motion_notify),
                ("button_press_event", self._on_hover_button_press),
            ]
        else:
            connect_pairs += [
                ("button_press_event", self._on_nonhover_button_press),
            ]
            if not self._multiple:
                connect_pairs.append(
                    ("motion_notify_event", self._on_nonhover_button_press))
        self._disconnectors = [
            partial(canvas.mpl_disconnect, canvas.mpl_connect(*pair))
            for pair in connect_pairs
            for canvas in {artist.figure.canvas for artist in artists}]

        bindings = {**_default_bindings,
                    **(bindings if bindings is not None else {})}
        unknown_bindings = {*bindings} - {*_default_bindings}
        if unknown_bindings:
            raise ValueError("Unknown binding(s): {}".format(
                ", ".join(sorted(unknown_bindings))))
        bindings_items = list(bindings.items())
        for i in range(len(bindings)):
            action, key = bindings_items[i]
            for j in range(i):
                other_action, other_key = bindings_items[j]
                if key == other_key and key is not None:
                    raise ValueError(
                        f"Duplicate bindings: {key} is used for "
                        f"{other_action} and for {action}")
        self.bindings = bindings

        self.annotation_kwargs = (
            annotation_kwargs if annotation_kwargs is not None
            else copy.deepcopy(_default_annotation_kwargs))
        self.annotation_positions = (
            annotation_positions if annotation_positions is not None
            else copy.deepcopy(_default_annotation_positions))
        self.highlight_kwargs = (
            highlight_kwargs if highlight_kwargs is not None
            else copy.deepcopy(_default_highlight_kwargs))

    @property
    def artists(self):
        """The tuple of selectable artists."""
        return tuple(filter(_is_alive, (ref() for ref in self._artists)))

    @property
    def enabled(self):
        """Whether clicks are registered for picking and unpicking events."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    @property
    def selections(self):
        r"""The tuple of current `Selection`\s."""
        for sel in self._selections:
            if sel.annotation.axes is None:
                raise RuntimeError("Annotation unexpectedly removed; "
                                   "use 'cursor.remove_selection' instead")
        return tuple(self._selections)

    @property
    def visible(self):
        """
        Whether selections are visible by default.

        Setting this property also updates the visibility status of current
        selections.
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        self._visible = value
        for sel in self.selections:
            sel.annotation.set_visible(value)
            sel.annotation.figure.canvas.draw_idle()

    def add_selection(self, pi, fig, ax):
        """
        Create an annotation for a `Selection` and register it.

        Returns a new `Selection`, that has been registered by the `Cursor`,
        with the added annotation set in the :attr:`annotation` field and, if
        applicable, the highlighting artist in the :attr:`extras` field.

        Emits the ``"add"`` event with the new `Selection` as argument.  When
        the event is emitted, the position of the annotation is temporarily
        set to ``(nan, nan)``; if this position is not explicitly set by a
        callback, then a suitable position will be automatically computed.

        Likewise, if the text alignment is not explicitly set but the position
        is, then a suitable alignment will be automatically computed.
        """
        # This method takes fig & ax as additional parameters to be robust when
        # picking a Container/sub-artist which may be missing .figure/.axes.
        # pi: "pick_info", i.e. an incomplete selection.
        get_cached_renderer = (
            fig.canvas.get_renderer
            if hasattr(fig.canvas, "get_renderer")
            else ax.get_renderer_cache)  # mpl<3.6.
        renderer = get_cached_renderer()
        if renderer is None:
            fig.canvas.draw()  # Needed below anyways.
            renderer = get_cached_renderer()
        ann = ax.annotate(
            "", xy=pi.target, xytext=(np.nan, np.nan),
            horizontalalignment=_MarkedStr("center"),
            verticalalignment=_MarkedStr("center"),
            visible=self.visible,
            zorder=np.inf,
            **self.annotation_kwargs)
        # Move the Annotation's ownership from the Axes to the Figure, so that
        # it gets drawn even above twinned axes.  But ann.axes must stay set,
        # so that e.g. unit converters get correctly applied.
        ann.remove()
        ann.axes = ax
        fig.add_artist(ann)
        ann.draggable(use_blit=not self._multiple)
        extras = []
        if self._highlight:
            hl = self.add_highlight(*pi)
            if hl:
                extras.append(hl)
        sel = pi._replace(annotation=ann, extras=extras)
        # Update the text after setting the annotation on the Selection, so
        # that get_ann_text can safely access sel.annotation.axes even if
        # artist.axes itself is unset (Containers/sub-artist picks).
        ann.set_text(_pick_info.get_ann_text(*sel))
        self._selections.append(sel)
        self._selection_stack.append(sel)
        for cb in self._callbacks["add"]:
            cb(sel)

        user_set_ha = not isinstance(ann.get_horizontalalignment(), _MarkedStr)
        user_set_va = not isinstance(ann.get_verticalalignment(), _MarkedStr)

        # Check that `ann.axes` is still set, as callbacks may have removed the
        # annotation.
        if ann.axes and ann.xyann == (np.nan, np.nan):
            fig_bbox = fig.get_window_extent()
            ax_bbox = ax.get_window_extent()
            overlaps = []
            for idx, annotation_position in enumerate(
                    self.annotation_positions):
                if user_set_ha:
                    annotation_position = {
                        k: v for k, v in annotation_position.items()
                        if k not in ["ha", "horizontalalignment"]}
                if user_set_va:
                    annotation_position = {
                        k: v for k, v in annotation_position.items()
                        if k not in ["va", "verticalalignment"]}
                ann.set(**annotation_position)
                # Work around matplotlib/matplotlib#7614: position update is
                # missing.
                ann.update_positions(renderer)
                bbox = ann.get_window_extent(renderer)
                overlaps.append(
                    (_get_rounded_intersection_area(fig_bbox, bbox),
                     _get_rounded_intersection_area(ax_bbox, bbox),
                     # Avoid needlessly jumping around by breaking ties using
                     # the last used position as default.
                     idx == self._last_auto_position))
            auto_position = max(range(len(overlaps)), key=overlaps.__getitem__)
            annotation_position = self.annotation_positions[auto_position]
            if user_set_ha:
                annotation_position = {
                    k: v for k, v in annotation_position.items()
                    if k not in ["ha", "horizontalalignment"]}
            if user_set_va:
                annotation_position = {
                    k: v for k, v in annotation_position.items()
                    if k not in ["va", "verticalalignment"]}
            ann.set(**annotation_position)
            self._last_auto_position = auto_position
        else:
            if not user_set_ha:
                ann.set_horizontalalignment(
                    {-1: "right", 0: "center", 1: "left"}[
                        np.sign(np.nan_to_num(ann.xyann[0]))])
            if not user_set_ha:
                ann.set_verticalalignment(
                    {-1: "top", 0: "center", 1: "bottom"}[
                        np.sign(np.nan_to_num(ann.xyann[1]))])

        if (extras
                or len(self.selections) > 1 and not self._multiple
                or not fig.canvas.supports_blit):
            # Either:
            #  - there may be more things to draw, or
            #  - annotation removal will make a full redraw necessary, or
            #  - blitting is not (yet) supported.
            fig.canvas.draw_idle()
        elif ann.axes:
            # Fast path, only needed if the annotation has not been immediately
            # removed.
            ann.draw(renderer)
            fig.canvas.blit()
        # Removal comes after addition so that the fast blitting path works.
        if not self._multiple:
            for other in self.selections[:-1]:
                self.remove_selection(other)
        return sel

    def add_highlight(self, artist, *args, **kwargs):
        """
        Create, add, and return a highlighting artist.

        This method is should be called with an "unpacked" `Selection`,
        possibly with some fields set to None.

        It is up to the caller to register the artist with the proper
        `Selection` (by calling ``sel.extras.append`` on the result of this
        method) in order to ensure cleanup upon deselection.
        """
        hl = _pick_info.make_highlight(
            artist, *args,
            **{"highlight_kwargs": self.highlight_kwargs, **kwargs})
        if hl:
            artist.axes.add_artist(hl)
            return hl

    def connect(self, event, func=None):
        """
        Connect a callback to a `Cursor` event; return the callback.

        Two events can be connected to:

        - callbacks connected to the ``"add"`` event are called when a
          `Selection` is added, with that selection as only argument;
        - callbacks connected to the ``"remove"`` event are called when a
          `Selection` is removed, with that selection as only argument.

        This method can also be used as a decorator::

            @cursor.connect("add")
            def on_add(sel):
                ...

        Examples of callbacks::

            # Change the annotation text and alignment:
            lambda sel: sel.annotation.set(
                text=sel.artist.get_label(),  # or use e.g. sel.index
                ha="center", va="bottom")

            # Make label non-draggable:
            lambda sel: sel.draggable(False)

        Note that when a single event causes both the removal of an "old"
        selection and the addition of a "new" one (typically, clicking on an
        artist when another one is selected, or hovering -- both assuming that
        ``multiple=False``), the "add" callback is called *first*.  This allows
        it, in particular, to "cancel" the addition (by immediately removing
        the "new" selection) and thus avoid removing the "old" selection.
        However, this call order may change in a future release.
        """
        if event not in self._callbacks:
            raise ValueError(f"{event!r} is not a valid cursor event")
        if func is None:
            return partial(self.connect, event)
        self._callbacks[event].append(func)
        return func

    def disconnect(self, event, cb):
        """
        Disconnect a previously connected callback.

        If a callback is connected multiple times, only one connection is
        removed.
        """
        try:
            self._callbacks[event].remove(cb)
        except KeyError:
            raise ValueError(f"{event!r} is not a valid cursor event")
        except ValueError:
            raise ValueError(f"Callback {cb} is not registered to {event}")

    def remove(self):
        """
        Remove a cursor.

        Remove all `Selection`\\s, disconnect all callbacks, and allow the
        cursor to be garbage collected.
        """
        for disconnector in self._disconnectors:
            disconnector()
        for sel in self.selections:
            self.remove_selection(sel)
        for s in type(self)._keep_alive.values():
            with suppress(KeyError):
                s.remove(self)

    def _on_pick(self, event):
        # Avoid creating a new annotation when dragging a preexisting
        # annotation (if multiple = True).  To do so, rely on the fact that
        # pick_events (which are used to implement dragging) trigger first (via
        # Figure's button_press_event, which is registered first); when one of
        # our annotations is picked, registed the corresponding mouse event as
        # "suppressed".  This can be done via a WeakSet as Matplotlib will keep
        # the event alive while being propagated through the callbacks.
        # Additionally, also rely on this mechanism to update the "current"
        # selection.
        for sel in self._selections:
            if event.artist is sel.annotation:
                self._suppressed_events.add(event.mouseevent)
                self._selection_stack.remove(sel)
                self._selection_stack.append(sel)
                break

    def _on_nonhover_button_press(self, event):
        if _mouse_event_matches(event, self.bindings["select"]):
            self._on_select_event(event)
        if _mouse_event_matches(event, self.bindings["deselect"]):
            self._on_deselect_event(event)

    def _on_hover_motion_notify(self, event):
        if event.button is None:
            # Filter away events where the mouse is pressed, in particular to
            # avoid conflicts between hover and draggable.
            self._on_select_event(event)

    def _on_hover_button_press(self, event):
        if _mouse_event_matches(event, self.bindings["deselect"]):
            # Still allow removing the annotation by right clicking.
            self._on_deselect_event(event)

    def _filter_mouse_event(self, event):
        # Accept the event iff we are enabled, and either
        # - no other widget is active, and this is not the second click of a
        #   double click (to prevent double selection), or
        # - another widget is active, and this is a double click (to bypass
        #   the widget lock), or
        # - hovering is active (in which case this is a motion_notify_event
        #   anyways).
        return (self.enabled
                and (event.canvas.widgetlock.locked() == event.dblclick
                     or self._hover))

    def _gen_sorted_candidates(self, event):
        per_axes_event = functools.lru_cache(None)(
            partial(_reassigned_axes_event, event))  # Fix twin axes support.
        pifas = []
        for artist in self.artists:
            if (artist.axes is None  # Removed or figure-level artist.
                    or event.canvas is not artist.figure.canvas
                    or not artist.get_visible()
                    or not artist.axes.contains(event)[0]):  # Cropped by axes.
                continue
            pi = _pick_info.compute_pick(artist, per_axes_event(artist.axes))
            if pi:
                pifas.append((pi, artist.figure, artist.axes))
        return sorted(pifas, key=lambda pifa: pifa[0].dist)

    def _on_select_event(self, event):
        if (not self._filter_mouse_event(event)
                # See _on_pick.  (We only suppress selects, not deselects.)
                or event in self._suppressed_events):
            return
        candidates = self._gen_sorted_candidates(event)
        # If the pick corresponds to an already selected artist at the same
        # point, the user is likely just dragging it.
        duplicates = {(other.artist, tuple(other.target))
                      for other in self._selections}
        candidates_nodupe = [
            (pi, fig, ax) for pi, fig, ax in candidates
            if (pi.artist, tuple(pi.target)) not in duplicates]
        if candidates_nodupe:
            return self.add_selection(*candidates_nodupe[0])
        elif not candidates and self._hover == HoverMode.Transient:
            # In transient hover mode, selections should be cleared if no
            # candidate picks *including duplicates* have been seen.
            for sel in self.selections:
                if event.canvas is sel.annotation.figure.canvas:
                    self.remove_selection(sel)

    def _on_deselect_event(self, event):
        if not self._filter_mouse_event(event):
            return
        for sel in self.selections[::-1]:  # LIFO.
            ann = sel.annotation
            if event.canvas is not ann.figure.canvas:
                continue
            if ann.contains(event)[0]:
                self.remove_selection(sel)
                break
        else:
            if self._highlight:
                for sel in self.selections[::-1]:
                    if any(extra.contains(event)[0] for extra in sel.extras):
                        self.remove_selection(sel)
                        break

    def _on_key_press(self, event):
        if event.key == self.bindings["toggle_enabled"]:
            self.enabled = not self.enabled
        elif event.key == self.bindings["toggle_visible"]:
            self.visible = not self.visible
        if not self._selections or not self.enabled:
            return
        sel = self._selection_stack[-1]
        for key in ["left", "right", "up", "down"]:
            if event.key == self.bindings[key]:
                self.add_selection(
                    _pick_info.move(*sel, key=key),
                    sel.annotation.figure, sel.annotation.axes)
                if self._multiple:  # Else, already removed.
                    self.remove_selection(sel)  # Also unsets .figure, .axes.
                break

    def select_at(self, target, xy):
        """
        Trigger and return a selection on *target* at data coordinates *xy*.

        *target* can be an axes or an artist.  The selection is guaranteed to
        be, in the first case, on a child artist; in the second case, on that
        specific artist.  If a click at *xy* does not result in a valid
        selection, then None is returned.

        Note
        ----
        This API is experimental and subject to future adjustments.
        """
        if isinstance(target, Axes):
            ax = target
        elif target in self.artists:
            ax = target.axes
        else:
            raise ValueError(f"Not a valid target: {target}")
        event = MouseEvent("button_press_event", ax.figure.canvas,
                           *ax.transData.transform(xy))
        candidates = [
            (pi, fig, ax) for pi, fig, ax in self._gen_sorted_candidates(event)
            if target in [pi.artist, ax]]
        if candidates:
            return self.add_selection(*candidates[0])

    def remove_selection(self, sel):
        """Remove a `Selection`."""
        self._selections.remove(sel)
        self._selection_stack.remove(sel)
        # <artist>.figure will be unset so we save them first.
        figs = {artist.figure for artist in [sel.annotation] + sel.extras}
        # ValueError is raised if the artist has already been removed.
        with suppress(ValueError):
            sel.annotation.remove()
        for artist in sel.extras:
            with suppress(ValueError):
                artist.remove()
        for cb in self._callbacks["remove"]:
            cb(sel)
        for fig in figs:
            fig.canvas.draw_idle()


def cursor(pickables=None, **kwargs):
    """
    Create a `Cursor` for a list of artists, containers, and axes.

    Parameters
    ----------

    pickables : Optional[List[Union[Artist, Container, Axes, Figure]]]
        All artists and containers in the list or on any of the axes or
        figures passed in the list are selectable by the constructed `Cursor`.
        Defaults to all artists and containers on any of the figures that
        :mod:`~matplotlib.pyplot` is tracking.  Note that the latter will only
        work when relying on pyplot, not when figures are directly instantiated
        (e.g., when manually embedding Matplotlib in a GUI toolkit).

    **kwargs
        Keyword arguments are passed to the `Cursor` constructor.
    """

    # Explicit check to avoid a confusing
    # "TypeError: Cursor.__init__() got multiple values for argument 'artists'"
    if "artists" in kwargs:
        raise TypeError(
            "cursor() got an unexpected keyword argument 'artists'")

    if pickables is None:
        # Do not import pyplot ourselves to avoid forcing the backend.
        plt = sys.modules.get("matplotlib.pyplot")
        pickables = [
            plt.figure(num) for num in plt.get_fignums()] if plt else []
    elif (isinstance(pickables, Container)
          or not isinstance(pickables, Iterable)):
        pickables = [pickables]

    def iter_unpack_figures(pickables):
        for entry in pickables:
            if isinstance(entry, Figure):
                yield from entry.axes
            else:
                yield entry

    def iter_unpack_axes(pickables):
        for entry in pickables:
            if isinstance(entry, Axes):
                yield from _iter_axes_subartists(entry)
                containers.extend(entry.containers)
            elif isinstance(entry, Container):
                containers.append(entry)
            else:
                yield entry

    containers = []
    artists = [*iter_unpack_axes(iter_unpack_figures(pickables))]
    for container in containers:
        contained = [*filter(None, container.get_children())]
        for artist in contained:
            with suppress(ValueError):
                artists.remove(artist)
        if contained:
            artists.append(_pick_info.ContainerArtist(container))

    return Cursor(artists, **kwargs)



"""
Cursors on images
=================

... display the underlying data value.
"""

from matplotlib.offsetbox import AnnotationBbox, OffsetImage
import matplotlib.pyplot as plt
import numpy as np
import mplcursors

data = np.arange(100).reshape((10, 10))

fig, axs = plt.subplots(ncols=2)
axs[0].imshow(data, origin="lower")
axs[1].imshow(data, origin="upper", extent=[2, 3, 4, 5])

axs[1].set(xlim=(2, 4), ylim=(4, 6))
axs[1].add_artist(AnnotationBbox(OffsetImage(data), (3.5, 5.5)))

mplcursors.cursor()

fig.suptitle("Click anywhere on the image")

plt.show()
